use esp_idf_sys::{self as _}; // If using the `binstart` feature of `esp-idf-sys`, always keep this module imported
use esp_idf_hal::{
    prelude::Peripherals, 
    gpio::{AnyInputPin, AnyOutputPin},
    delay::BLOCK,
    units::Hertz,
    uart,
    i2c::{I2cConfig, I2cDriver} 
};
use esp_idf_hal;

use embedded_graphics::{
    prelude::*,
    image::{Image, ImageRaw},
    mono_font::{ascii::FONT_7X13, MonoTextStyle},
    pixelcolor::BinaryColor,
    text::{Text, Alignment},
};

use ssd1306::{prelude::*, I2CDisplayInterface, Ssd1306};

mod error;
use error::Error;

mod gps;
use gps::gpsdata;

mod display;
use display::DisplayError;

fn main() -> Result<(), Error> {
    // It is necessary to call this function once. Otherwise some patches to the runtime
    // implemented by esp-idf-sys might not link properly. See https://github.com/esp-rs/esp-idf-template/issues/71
    esp_idf_sys::link_patches();
    // Bind the log crate to the ESP Logging facilities
    esp_idf_svc::log::EspLogger::initialize_default();

    // Get Peripherals and sysloop ready
    let peripherals = Peripherals::take().unwrap();
    let pins = peripherals.pins;

    // Define program's pins + addresses
    let scl = pins.gpio22;
    let sda = pins.gpio21;
    let gps_rx  = pins.gpio16;
    let _lcd_address: u8 = 0x3C;

    
    // Setup UART to read serial data from GPS
    let config = uart::config::Config::default().baudrate(Hertz(9_600));
    let uart: uart::UartRxDriver = uart::UartRxDriver::new(
        peripherals.uart1,
        gps_rx,
        None::<AnyInputPin>,
        None::<AnyOutputPin>,
        &config
    ).map_err(|err| Error::EspError(err))?;

    // Setup I2C driver
    let config = I2cConfig::new().baudrate(Hertz(921600));
    let i2c = I2cDriver::new(
        peripherals.i2c0, 
        sda, 
        scl, 
        &config
    ).map_err(|err| Error::EspError(err))?;

    let interface = I2CDisplayInterface::new(i2c);
    let mut display = Ssd1306::new(interface, DisplaySize128x64, DisplayRotation::Rotate0)
        .into_buffered_graphics_mode();
    display.init().unwrap();

    let raw: ImageRaw<BinaryColor> = ImageRaw::new(include_bytes!("./rust.raw"), 64);
    let im: Image<'_, _> = Image::new(&raw, Point::new(32, 0));

    im.draw(&mut display).map_err(|err| Error::DisplayError(DisplayError::SetupError(err)))?;

    let mut latest_data = gpsdata::GpsData::default();
    
    loop {
        // Read buffer from UART
        let mut buf: Vec<u8> = (
            0..uart.count().map_err(|err| Error::EspError(err))? as u8
        ).collect();
        uart.read(&mut buf[..], BLOCK).map_err(|err| Error::EspError(err))?;

        //Update GPS Data Struct with the latest data fetched from UART buffer
        latest_data.update(&buf)
        .map_err(|err| Error::GpsError(err))?;

        // Clear display ready for next write
        let style = MonoTextStyle::new(&FONT_7X13, BinaryColor::On);
        display.clear(BinaryColor::Off)
        .map_err(|_| Error::DisplayError(DisplayError::DrawingError))?;

        // Draw text to Display
        Text::with_alignment(
            &latest_data.to_string().as_str(),
            Point::new(64, 10),
            style,
            Alignment::Center,
        )
        .draw(&mut display)
        .map_err(|_| Error::DisplayError(DisplayError::DrawingError))?;

        //Flush data to the display
        display.flush()
        .map_err(|_| Error::DisplayError(DisplayError::FlushError))?;
    }
}